@@ -40,6 +40,9 @@ sslBuffer_Grow(sslBuffer *b, unsigned int newLen)
         return SECSuccess;
     }
 
+    /* If buf is non-NULL, space must be non-zero;
+     * if buf is NULL, space must be zero. */
+    PORT_Assert((b->buf && b->space) || (!b->buf && !b->space));
     newLen = PR_MAX(newLen, b->len + 1024);
     if (newLen > b->space) {
         unsigned char *newBuf;
@@ -57,6 +60,22 @@ sslBuffer_Grow(sslBuffer *b, unsigned int newLen)
     return SECSuccess;
 }
 
+/* Appends len copies of c to b */
+SECStatus
+sslBuffer_Fill(sslBuffer *b, PRUint8 c, size_t len)
+{
+    PORT_Assert(b);
+    SECStatus rv = sslBuffer_Grow(b, b->len + len);
+    if (rv != SECSuccess) {
+        return SECFailure;
+    }
+    if (len > 0) {
+        memset(SSL_BUFFER_NEXT(b), c, len);
+    }
+    b->len += len;
+    return SECSuccess;
+}
+
 SECStatus
 sslBuffer_Append(sslBuffer *b, const void *data, unsigned int len)
 {
@@ -323,12 +342,23 @@ ssl3_AppendHandshake(sslSocket *ss, const void *void_src, unsigned int bytes)
 }
 
 SECStatus
-ssl3_AppendHandshakeNumber(sslSocket *ss, PRUint64 num, unsigned int lenSize)
+ssl3_AppendHandshakeNumberSuppressHash(sslSocket *ss, PRUint64 num, unsigned int lenSize, PRBool suppressHash)
 {
+    if ((lenSize > 8) || ((lenSize < 8) && (num >= (1ULL << (8 * lenSize))))) {
+        PORT_SetError(SEC_ERROR_INVALID_ARGS);
+        return SECFailure;
+    }
+
     PRUint8 b[sizeof(num)];
     SSL_TRC(60, ("%d: number:", SSL_GETPID()));
     ssl_EncodeUintX(b, num, lenSize);
-    return ssl3_AppendHandshake(ss, b, lenSize);
+    return ssl3_AppendHandshakeInternal(ss, b, lenSize, suppressHash);
+}
+
+SECStatus
+ssl3_AppendHandshakeNumber(sslSocket *ss, PRUint64 num, unsigned int lenSize)
+{
+    return ssl3_AppendHandshakeNumberSuppressHash(ss, num, lenSize, PR_FALSE);
 }
 
 SECStatus
@@ -362,3 +392,43 @@ ssl3_AppendBufferToHandshakeVariable(sslSocket *ss, sslBuffer *buf,
 {
     return ssl3_AppendHandshakeVariable(ss, buf->buf, buf->len, lenSize);
 }
+
+SECStatus
+ssl3_MaybeUpdateHashWithSavedRecord(sslSocket *ss)
+{
+    SECStatus rv;
+    /* dtls13ClientMessageBuffer is not empty if ClientHello has sent DTLS1.3 */
+    if (ss->ssl3.hs.dtls13ClientMessageBuffer.len == 0) {
+        return SECSuccess;
+    }
+
+    size_t offset = 0;
+
+    /* the first clause checks the version that was received in ServerHello:
+     * only if it's DTLS1.3, we remove the necessary fields. 
+     * the second clause checks if we send 0rtt (see TestTls13ZeroRttDowngrade).
+     */
+    if ((ss->version == ss->ssl3.cwSpec->version || ss->ssl3.hs.zeroRttState == ssl_0rtt_sent)) {
+        if (ss->ssl3.hs.dtls13ClientMessageBuffer.len < 12) {
+            PORT_SetError(SEC_ERROR_INVALID_ARGS);
+            return SECFailure;
+        }
+
+        rv = ssl3_UpdateHandshakeHashes(ss, ss->ssl3.hs.dtls13ClientMessageBuffer.buf, 4);
+        if (rv != SECSuccess) {
+            return SECFailure;
+        }
+        offset = 12;
+    }
+
+    PORT_Assert(offset < ss->ssl3.hs.dtls13ClientMessageBuffer.len);
+    rv = ssl3_UpdateHandshakeHashes(ss, ss->ssl3.hs.dtls13ClientMessageBuffer.buf + offset,
+                                    ss->ssl3.hs.dtls13ClientMessageBuffer.len - offset);
+    if (rv != SECSuccess) {
+        return SECFailure;
+    }
+
+    sslBuffer_Clear(&ss->ssl3.hs.dtls13ClientMessageBuffer);
+    ss->ssl3.hs.dtls13ClientMessageBuffer.len = 0;
+    return SECSuccess;
+}
